"""
LAMMPS tools
============
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides a number of extensions to the LAMMPS simulation
toolkit.
"""

from . import topology

__all__ = ["topology"]
