"""
Plotting settings and tools
===========================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains Matplotlib rcParams and tools for generating
aesthetically-pleasing scientific figures.
"""

from . import axis, color, rcparam

__all__ = ["axis", "color", "rcparam"]
